//+------------------------------------------------------------------+
//|                                            Linear_Regression.mq5 |
//|                                                         forexmts |
//|                                                  forexmt@mail.ru |
//+------------------------------------------------------------------+
#property copyright "Juggernaut"
#property link      "https://www.mql5.com/en/users/46961250"
#property version   "1.00"
#property indicator_chart_window
#property indicator_buffers 3
#property indicator_plots   3
//--- plot Basis
#property indicator_label1  "Basis"
#property indicator_type1   DRAW_LINE
#property indicator_color1  clrPurple
#property indicator_style1  STYLE_SOLID
#property indicator_width1  1
//--- plot Upper
#property indicator_label2  "Upper"
#property indicator_type2   DRAW_LINE
#property indicator_color2  clrPurple
#property indicator_style2  STYLE_SOLID
#property indicator_width2  1
//--- plot Lower
#property indicator_label3  "Lower"
#property indicator_type3   DRAW_LINE
#property indicator_color3  clrPurple
#property indicator_style3  STYLE_SOLID
#property indicator_width3  1

//--- input parameters
input int      iRegPeriod = 40;                             //Linear Regression Period
input double   dRegChannel = 2.0;                           //Linear Regression Channel
input color    clRegBasisColor = clrPurple;                 //Basis Line color
input ENUM_LINE_STYLE enRegBasisStyle = STYLE_DOT;          //Basis Line style
input int iRegBasisWidth = 1;                               //Basis Line width
input color    clRegChannelColor = clrPurple;               //Channel Line color
input ENUM_LINE_STYLE enRegChannelStyle = STYLE_SOLID;      //Channel Line style
input int iRegChannelWidth = 2;                             //Channel Line width


//--- indicator buffers
double         dBasisBuffer[];
double         dUpperBuffer[];
double         dLowerBuffer[];

//--- global variables
struct STRUCT_REGR
  {
  double a;
  double b;
  int bar1;
  double price1;
  int bar2;
  double price2;
  };
STRUCT_REGR stRegr;
//+------------------------------------------------------------------+
//| Custom indicator initialization function                         |
//+------------------------------------------------------------------+
int OnInit()
  {
  //--- entry input check
  if(iRegPeriod <= 0) { Alert("Error: 'Linear Regression Period' must be positive value!"); return(INIT_FAILED); }
  
  //--- indicator buffers mapping
  SetIndexBuffer(0, dBasisBuffer, INDICATOR_DATA);
  SetIndexBuffer(1, dUpperBuffer, INDICATOR_DATA);
  SetIndexBuffer(2, dLowerBuffer, INDICATOR_DATA);
  
  PlotIndexSetDouble(0, PLOT_EMPTY_VALUE, 0.0);
  PlotIndexSetDouble(1, PLOT_EMPTY_VALUE, 0.0);
  PlotIndexSetDouble(2, PLOT_EMPTY_VALUE, 0.0);
  
  PlotIndexSetInteger(0, PLOT_LINE_COLOR, clRegBasisColor);
  PlotIndexSetInteger(0, PLOT_LINE_STYLE, enRegBasisStyle);
  PlotIndexSetInteger(0, PLOT_LINE_WIDTH, iRegBasisWidth);
  PlotIndexSetInteger(1, PLOT_LINE_COLOR, clRegChannelColor);
  PlotIndexSetInteger(1, PLOT_LINE_STYLE, enRegChannelStyle);
  PlotIndexSetInteger(1, PLOT_LINE_WIDTH, iRegChannelWidth);
  PlotIndexSetInteger(2, PLOT_LINE_COLOR, clRegChannelColor);
  PlotIndexSetInteger(2, PLOT_LINE_STYLE, enRegChannelStyle);
  PlotIndexSetInteger(2, PLOT_LINE_WIDTH, iRegChannelWidth);
  
  //--- indicator settings
  IndicatorSetString(INDICATOR_SHORTNAME, StringFormat("LinReg(%i,%.1f)", iRegPeriod, dRegChannel));
  
  return(INIT_SUCCEEDED);
  }
//+------------------------------------------------------------------+
//| Custom indicator iteration function                              |
//+------------------------------------------------------------------+
int OnCalculate(const int rates_total,
                const int prev_calculated,
                const datetime &time[],
                const double &open[],
                const double &high[],
                const double &low[],
                const double &close[],
                const long &tick_volume[],
                const long &volume[],
                const int &spread[])
  {
  //--- not enough bars
  if(rates_total < iRegPeriod) return(0);
  
  //--- reset buffer values
  ZeroMemory(dBasisBuffer);
  ZeroMemory(dUpperBuffer);
  ZeroMemory(dLowerBuffer);
  
  //linear regression calculation
  stRegr.bar2=rates_total-1;
  stRegr.bar1=stRegr.bar2-iRegPeriod+1;
  int n=stRegr.bar2-stRegr.bar1+1;
  
  //---- calculate price values
  double value=close[stRegr.bar2];
  double c;
  double sumy=value;
  double sumx=0.0;
  double sumxy=0.0;
  double sumx2=0.0;
  for(int i=1; i<n; i++)
    {
    value=close[stRegr.bar2-i];
    sumy+=value;
    sumxy+=value*i;
    sumx+=i;
    sumx2+=i*i;
    }
  c=sumx2*n-sumx*sumx;
  if(c==0.0) return(0);
  stRegr.b=(sumxy*n-sumx*sumy)/c;
  stRegr.a=(sumy-sumx*stRegr.b)/n;
  stRegr.price2=stRegr.a;
  stRegr.price1=stRegr.a+stRegr.b*n;
  
  //---- maximal deviation calculation
  double maxdev=0;
  double deviation=0;
  double dvalue=stRegr.a;
  for(int i=0; i<n; i++)
    {
    value=close[stRegr.bar2-i];
    dvalue+=stRegr.b;
    deviation=MathAbs(value-dvalue);
    if(maxdev<=deviation) maxdev=deviation;
    }
  
  //--- assign buffers
  for(int bar = stRegr.bar2, i = 0; bar >= stRegr.bar1; bar--, i++)
    {
    dBasisBuffer[bar] = stRegr.a + i * stRegr.b;
    dUpperBuffer[bar] = dBasisBuffer[bar] + maxdev * dRegChannel;
    dLowerBuffer[bar] = dBasisBuffer[bar] - maxdev * dRegChannel;
    }
    
  //--- return value of prev_calculated for next call
  return(rates_total);
  }
//+------------------------------------------------------------------+


